/* Jon Arney, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#else
#include <io.h>
#endif

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <stdarg.h>

#include "gnut_lib.h"
#include "gnut_connection.h"
#include "gnut_net.h"
#include "host.h"
#include "gnut_transfer.h"
#include "query.h"
#include "conf.h"
#include "share.h"
#include "gnut_threads.h"
#include "gnut_http.h"
#include "Gnut_Hash.h"

#define GNUT_HTTP_CACHE_EXPIRE 20

#ifdef GNUT_HTTP_SEARCH
int cgi_parse( char * CGI, char * field, char * value );
static void gnut_guid_to_string(char *string, char *guid);
//static void gnut_string_to_guid(char *guid, char *string);

Gnut_Hash *query_cache;

char *html_template_head;
char *html_template_foot;

static uchar gnut_http_query_cache_hash(void *p_data)
{
	gnut_http_query_t *data;
	uchar hashval;
	int i;

	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");

	data = (gnut_http_query_t *)p_data;

	hashval = 0;
	for (i = 1; i < 16; i++) {
	  if (i != 6) {
		hashval += data->guid[i];
	  }
	}
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT\n");
	return hashval;
}

static int gnut_http_query_cache_compare(void *p_a, void *p_b)
{
	gnut_http_query_t *a;
	gnut_http_query_t *b;
	char stringa[65];
	char stringb[65];

	a = (gnut_http_query_t *)p_a;
	b = (gnut_http_query_t *)p_b;

	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");

	/*
	 * They match if the guid's match.
	 */
	gnut_guid_to_string(stringa, a->guid);
	gnut_guid_to_string(stringb, b->guid);
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Hasha = %s\n", stringa);
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Hashb = %s\n", stringa);
	a->guid[6] = '\0';
	b->guid[6] = '\0';
	if (memcmp(a->guid, b->guid, sizeof(a->guid)) == 0) {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - Found an equal hash by GUID\n");
	  return 0;
	}

	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - Hash not equal\n");
	return -1;
}

static void gnut_http_query_cleanup(void)
{
  gnut_http_query_t *req;
  Gnut_List *list;
  int i;
  time_t last_time_to_keep;

  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");

  if (!query_cache) {
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - nothing to clean\n");
    return;
  }

  /*
   * Do garbage collection on the hash.
   * based on 'oldest'.  Anything older than
   * EXPIRE minutes gets deleted.
   */
  last_time_to_keep = time(NULL);
  last_time_to_keep -= (GNUT_HTTP_CACHE_EXPIRE*60);

  /*
   * For each element of the hash,
   * remove the oldest.
   */
  for (i = 0; i < 256; i++) {
    list = query_cache->list[i];

    /*
     * Walk the list looking for
     * old guys, and delete them.
     */
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Starting garbage collection with %i\n",
    	    gnut_list_size(query_cache->list[i]));

    while (1) {
      if (!list) break;
      req = list->data;
      if (!req) break;

      /*
       * This is a hack because
       * I'm trying to remove
       * an entry from the same
       * list I'm walking.
       */
      if (req->last_used < last_time_to_keep) {
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Removing an old req from the list\n");
	list = gnut_list_remove(list, req);
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Freeing an old one\n");
	xfree(req);
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Re-assigning the list\n");
	query_cache->list[i] = list;
      }
      else {
	list = gnut_list_next(list);
      }
    }
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Finished garbage collection, now have %i\n",
	    gnut_list_size(query_cache->list[i]));
  }
  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT\n");
}

Gnut_List *gnut_http_query_get(gnut_transfer *gt, char *guid)
{
  gnut_http_query_t *req, req_findme;

  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");

  memcpy(req_findme.guid, guid, 16);
  if (!query_cache) {
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no cache yet\n");
    return NULL;
  }
  req = gnut_hash_find(query_cache, &req_findme);
  if (!req) {
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no entry found in cache\n");
    return NULL;
  }

  req->last_used = time(NULL);

  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT\n");

  return req->query_list;
}

void gnut_http_query_new(gnut_transfer *gt, char *request, char *guid)
{
  gnut_http_query_t *req, req_findme;
  gnutella_packet *gpa;
  int i;

  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");

  if (request == NULL) {
    /*
     * The request was a frivolous one, ignore it.
     */
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - the query was frivolous\n");
    return;
  }

  gnut_http_query_cleanup();

  /*
   * Actually submit the request with the new MAC
   * and stuff.
   */
  gpa=gp_request_make(conf_get_str("mac"),request,conf_get_int("ttl"));
  for (i = 7; i < 16; i++) {
    gpa->gh.guid[i] = guid[i];
  }
  send_to_all(gpa);

  if (!query_cache) {
    query_cache = gnut_hash_new(gnut_http_query_cache_hash,
				gnut_http_query_cache_compare);
  }

  /*
   * See if we're already on the list....
   */
  //  memcpy(req_findme.ip, gt->ip, sizeof(gt->ip));
  //  req_findme.port = gt->port;
  memcpy(req_findme.guid, gpa->gh.guid, 16);
  req = gnut_hash_find(query_cache, &req_findme);
	
  /*
   * We're not on the list, so add us to the list.
   */
  if (!req) {
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Appending an http to the hash list\n");
    req = (gnut_http_query_t *)xmalloc(sizeof(gnut_http_query_t));
    memcpy(req->guid, gpa->gh.guid, 16);
    memcpy(req->ip, gt->ip, 4);
    req->port = gt->port;
    req->last_used = time(NULL);
    req->query_list = NULL;
    gnut_hash_insert(query_cache, req);
  }
  else {
    memcpy(req->guid, gpa->gh.guid, 16);
    req->last_used = time(NULL);
    g_debug(GNUT_HTTP_INTERNALS_DEBUG, "Updating the GUID to the hash list\n");
  }
  xfree(gpa->data);
  xfree(gpa);

  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT\n");
}

int gnut_http_result_append(
			     uchar ip[4],
			     uchar port[2],
			     int ref,
			     int speed,
			     int size,
			     char *name,
			     uchar guid[16]
			     )
{
	gnut_http_query_t *req, req_findme;
	query_response *qr;
	char string[64];


	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");
	if (!query_cache) {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no query to append result to\n");
	  return 0;
	}

	memcpy(req_findme.guid, guid, sizeof(req_findme.guid));
	gnut_guid_to_string(string, guid);
	req = gnut_hash_find(query_cache, &req_findme);
	if (!req) {
	  /*
	   * I doesn't appear to be our packet.
	   */
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - this query did not belong to us : %s\n", string);
	  return 0;
	}

	qr=(query_response *) xmalloc(sizeof(query_response));
	memcpy(qr->ip,ip,4);
	memcpy(qr->port,port,2);
	qr->ref=ref;
	qr->speed=speed;
	qr->size=size;
	qr->name=strdup(name);
	memcpy(qr->guid,guid,16);
	req->last_used = time(NULL);
	req->query_list = gnut_list_prepend(req->query_list , qr);

	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - this reply belonged to us : %s\n", string);
	return 1;
}

int gnut_http_file_search(gnut_transfer *gt, http_request_t *req)
{
	Gnut_List *list;
	query_response *qr;
	char *search_string = NULL;
	char *query_string;
	int ret;
	char c_port[32];
	char c_address[512];
	unsigned int i_port;
	char value[512];
#ifdef GNUT_HTTP_WAIT_FOR_RESULTS
	char c_gnut_min_returns[32];
	char c_gnut_time_to_wait[32];
	unsigned int gnut_time_to_wait;
	unsigned int gnut_min_returns;
	int n;
	int count;
#endif
	int i;
	char guid[16];
	http_cookie_t *cookp;

	g_debug(GNUT_HTTP_DEBUG, "IN\n");

	/*
	 * Check to see if we got a GUID in a cookie already...
	 */
	cookp = http_cookie_new();
	/*
	ret = http_header_get_cookie(req, cookp);
	if (ret) {
	*/
	  /*
	   * We had no cookie, so we will create a new guid
	   * and set it as a cookie to the browser.
	   */
	  g_debug(1,"creating new cookie\n");

	  cookp->value = xmalloc(sizeof(char)*33);

	  memcpy(guid,conf_get_str("mac"),6);
	  for (i=6;i<16;i++) guid[i]=rand();

	  gnut_guid_to_string(cookp->value, guid);
	  http_response_set_cookie(req, cookp);
	  g_debug(GNUT_HTTP_DEBUG, "Setting the guid in a cookie %s\n", cookp->value);
	  /*
	}
	else {
	  */
	  /*
	   * We got a cookie, use this for the GUID.
	   */
	  /*
	  gnut_string_to_guid(guid, cookp->value);
	  g_debug(GNUT_HTTP_DEBUG, "Got the guid in a cookie %s\n", cookp->value);
	}
	  */
	http_cookie_delete(cookp);
	
	/*
	 * Get the CGI string out of the query.
	 */
	query_string = strstr(req->url_file, "/search/?");
	if (query_string) {
	  query_string += 9;
	  ret = cgi_parse(query_string, "query", value);

	  g_debug(GNUT_HTTP_DEBUG, "Found %i:%s\n", ret, value);
	  if (!ret) {
	    search_string = value;
	  }

#ifdef GNUT_HTTP_WAIT_FOR_RESULTS
	  ret = cgi_parse(query_string, "returns", c_gnut_min_returns);

	  if (ret) {
	    gnut_min_returns = 20;
	  }
	  else {
	    gnut_min_returns = atoi(c_gnut_min_returns);
	  }
	  if (gnut_min_returns > 1200) gnut_min_returns = 1200;
	  g_debug(GNUT_HTTP_DEBUG, "Min Returns: %i\n", gnut_min_returns);

	  ret = cgi_parse(query_string, "wait", c_gnut_time_to_wait);
	  if (ret) {
	    gnut_time_to_wait = 3;
	  }
	  else {
	    gnut_time_to_wait = atoi(c_gnut_time_to_wait);
	  }
	  if (gnut_time_to_wait > 60) gnut_time_to_wait = 60;
	  g_debug(GNUT_HTTP_DEBUG, "Max time to wait for returns %i\n", gnut_time_to_wait);
#endif
	}

	/*
	 * This is old and should go away soon.
	 */
	else {
	  search_string = strstr(req->url_file, "/search/");
	  /*
	   * This should never happen.
	   */
	  if (!search_string) {
	    g_debug(GNUT_HTTP_DEBUG, "OUT - This is not really a query string\n");
	    return -1; 
	  }
	  search_string += 8;

#ifdef GNUT_HTTP_WAIT_FOR_RESULTS
	  gnut_min_returns = 20;
	  gnut_time_to_wait = 3;
#endif
	}

	g_debug(GNUT_HTTP_DEBUG, "Writing headers\n");

	http_response_header_add(req, "Content-Type", "text/html");
	http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
	http_response_header_add(req, "Connection", "close");
	http_response_header_write(req);

	g_debug(GNUT_HTTP_DEBUG, "Writing search form\n");

	gnut_http_write_top(req);
	gnut_http_write_form(req);

	/*
	 * End this now if there's no meaningful search to perform.
	 */

	if (search_string && strlen(search_string) >= 4) {
	  g_debug(GNUT_HTTP_DEBUG, "Starting new Gnutella-net query for %s\n", search_string);

	  gnut_http_query_new(gt, search_string, guid);

	  http_write_va(req, 512, "<h1>Search Results for %s</h1>", search_string);
	  http_write_str(req, "<ol>\r\n");
	  
	  /*
	   * Wait for  time_to_wait seconds,
	   * or min_returns requests, whichever
	   * is first. - By default, this is not a good
	   * thing to do, as it's a little cumbersome.
	   */
#ifdef GNUT_HTTP_WAIT_FOR_RESULTS
	  n = 0;
	  count = 0;
	  while (n < gnut_min_returns) {
	    //	    n = query_num();
	    //	    if (n >= gnut_min_returns) break;
	    if (count > gnut_time_to_wait) break;
	    count++;
	    sleep(1);
	  }
#endif	  
	}

	list = gnut_http_query_get(gt, guid);

	while(1) {
	  if (!list) break;
	  qr = list->data;
	  if (!qr) break;
	  http_write_str(req, "<li>");
	  
	  i_port = qr->port[1];
	  i_port <<= 8;
	  i_port |= qr->port[0];
	  sprintf(c_port, "%i", i_port);
	  sprintf(c_address, "%i.%i.%i.%i",
		  qr->ip[0],qr->ip[1],qr->ip[2],qr->ip[3]);
	  
	  http_write_va(req, 
			512,
			"<a href=\"http://%s:%s/get/%i/%s\">",
			c_address,
			c_port,
			qr->ref,
			qr->name);
	  http_write_str(req, qr->name);
	  http_write_str(req, "</a>");
	  http_write_str(req, "</li>\r\n");
	  list = gnut_list_next(list);
	  if (!list) break;
	}
	
	gnut_http_write_bottom(req);	
	return 0;
}
#endif

#ifdef GNUT_HTTP_FILE_LIST
int gnut_http_file_list(gnut_transfer *gt, http_request_t *req)
{
	Gnut_List *list;
	share_item *si;

	g_debug(GNUT_HTTP_DEBUG, "IN\n");

	g_debug(GNUT_HTTP_DEBUG, "Writing headers\n");
	http_response_header_add(req, "Content-Type", "text/html");
	http_response_header_add_va(req, "Server", "Gnut/%s", VERSION);
	http_response_header_add(req, "Connection", "close");
	http_response_header_write(req);

	g_debug(GNUT_HTTP_DEBUG, "Writing top\n");
	gnut_http_write_top(req);

	http_write_str(req, "<h2>Files on this server</h2>");
	http_write_str(req, "<table type=list>\n");

	g_debug(GNUT_HTTP_DEBUG, "Getting share root\n");
	list = share_get_root();

	g_debug(GNUT_HTTP_DEBUG, "Writing local share list\n");
	while(1) {
		if (!list) break;
		si = list->data;
		if (!si) break;

		http_write_str(req, "<tr><td>");
		http_write_va(req, 512, "<a href=\"/get/%i/%s\">", si->ref, si->path);
		http_write_str(req, si->path);
		http_write_str(req, "</a></td>");
		http_write_va(req, 512, "<td>%i</td>", si->size);
		http_write_str(req, "</tr>\n");

		list = gnut_list_next(list);
	}
	http_write_str(req, "</table>");

	g_debug(GNUT_HTTP_DEBUG, "Writing bottom of HTML page.\n");
	gnut_http_write_bottom(req);

	g_debug(GNUT_HTTP_DEBUG, "OUT\n");
	return 0;
}
#endif

static int parse_http_request(char *request, int *ref, char **name)
{
	int iref;
	char *ptr,*ptr2;
	
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN - %s\n", request);
	
	ptr=strstr(request,"/get/");
	if (ptr==NULL) {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no /get/ in url\n");
	  return -2;
	}

	/*
	 * Danger will robinson.
	 * Unsafe pointer manipulations.
	 */	
	ptr+=5;
	ptr2=strchr(ptr,'/');
	if (ptr2==NULL) {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no reference\n");
	  return -3;
	}
	
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "located end slash ptr = %s\n", ptr);
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "located end slash ptr2 = %s\n", ptr2);
	//	*ptr2=0;
	
	iref=atoi(ptr);
	
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "located ref num: %i\n",iref);
	
	//ptr2++;
	//	ptr = ptr2 + strlen(ptr2)-1;
	
	if (!(*ptr)) {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - no filename\n");
	  return -4;
	}
	
	*name=strdup(ptr2);
	*ref=iref;
	
	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT\n");
	return 0;
}

static int parse_range(char *buf, int size, int *rangemin, int *rangemax)
{
	int i;
	int num,num2;
	char *ptr;

	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "IN\n");
	
	for (i=6;i<strlen(buf) && !isdigit(buf[i]) && buf[i]!='-';i++);
	
	if (i==strlen(buf)) {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - full length, aparently\n");
	  return 0;
	}
	
	if (buf[i]=='-') {
		num=atoi(&buf[i+1]);
		*rangemin=size-num;
		*rangemax=size-1;
		g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - 1 - min=%i, max=%i\n", *rangemin, *rangemax);
		return 1;
	}
	
	num=(int) strtol(&buf[i],&ptr,10);
	
	if ((*ptr)!='-') {
	  g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - full length, aparently - bad range\n");
	  return 0;
	}
	
	ptr++;
	
	if (!isdigit(*ptr)) {
		*rangemin=num;
		*rangemax=size;
		g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - 2 - min=%i, max=%i\n", *rangemin, *rangemax);
		return 1;
	}
	
	num2=(int) strtol(ptr,NULL,10);
	
	*rangemin=num;
	*rangemax=num2;

	g_debug(GNUT_HTTP_INTERNALS_DEBUG, "OUT - 3 - min=%i, max=%i\n", *rangemin, *rangemax);
	return 1;
}

int gnut_http_serve_file(gnut_transfer *gt, http_request_t *req)
{
	int ref;
	int ret;
	share_item *si=NULL;
	char *buf=NULL;
	int partial=0,rangemin=0,rangemax=0;
	char *range;

	g_debug(GNUT_HTTP_DEBUG, "IN\n");
	range = http_request_find_header(req, "Range");

	/*
	 * Parse to find the actual file we're looking for.
	 */
	ret = parse_http_request(req->url_file, &ref, &gt->fname);
	g_debug(GNUT_HTTP_DEBUG, "parse_http returned %i ref=%i name=%s\n",ret, ref,gt->fname);

	if (ref) si=share_find(ref);

	// this next bit is not very intuitive, but I didn't want to have
	// to have 3 separate instances of cleanup code...
	if (ret<0 || si==NULL) {
		g_debug(GNUT_HTTP_DEBUG,"error with something!\n");
		if (si) {
			xfree(si->path);
			xfree(si->upath);
			xfree(si->fpath);
			xfree(si);
		}
		
		return -1;
	}

	buf=xmalloc(4096);

	/*
	 * If this was a partial content request,
	 * then parse the amount of the partial content.
	 */
	if (range) {
	  	partial = parse_range(range, si->size, &rangemin, &rangemax);
	  	g_debug(3,"range min: %i max: %i\n", rangemin, rangemax);
		// workaround Gnutella asking for a rangemin > file size 
		if (partial && (rangemin>rangemax)) partial=0;
	}

	gt->fsock=open(si->fpath,O_RDONLY);
	
	if (gt->fsock<0) {
		g_debug(GNUT_HTTP_DEBUG,"couldn't open something!\n");
		
		if (gt->fsock>=0) close(gt->fsock);
		xfree(buf);
		if (si) {
			xfree(si->path);
			xfree(si->upath);
			xfree(si->fpath);
			xfree(si);
		}

		return -1;
	}
	
	
	
	g_debug(GNUT_HTTP_DEBUG,"writing out http reply\n");

	if (partial==0) {
		gt->total=si->size;
		sprintf(buf, "Gnut/%s", VERSION);
		http_response_header_add(req, "Server", buf);
		http_response_header_add(req, "Content-Type", "application/octet-stream");
		http_response_header_add_va(req, "Content-Length", "%i", si->size);
		http_response_header_write(req);
	} else {
		gt->bytes=rangemin;
		gt->total=rangemax+1;

		// I used to have the correct header here
		// HTTP 206 Partial Content
		// but the real gnutella thought this was an error
		// so I changed it

		http_response_header_add(req,
					 "Content-Type",
					 "application/octet-stream");
		http_response_header_add_va(
					    req,
					    "Content-Range",
					    "bytes %i-%i",
					    rangemin, rangemax);
		http_response_header_add_va(
					    req,
					    "Content-Length",
					    "%i", 
					    rangemax - rangemin + 1);
		http_response_header_write(req);

		lseek(gt->fsock,rangemin,SEEK_SET);
	}

	
	gt->state=STATE_CONNECTED;
	
	gnut_transfer_loop(gt);
	
	g_debug(GNUT_HTTP_DEBUG,"closing up transfer\n");
		
	xfree(buf);

	g_debug(GNUT_HTTP_DEBUG, "OUT\n");
	return 0;
}

#if GNUT_HTTP_FILE_LIST || GNUT_HTTP_SEARCH
void gnut_http_write_top(http_request_t *req)
{
	if (html_template_head) {
		http_write_str(req, html_template_head);
	} else {
		http_write_str(req, "<HTML>\r\n");
		http_write_str(req, "<HEAD><TITLE>GNUT file sharer</TITLE></HEAD>\r\n");
		http_write_str(req, "<BODY BGCOLOR=\"#FFFFFF\">\r\n");
		http_write_str(req, "<A HREF=\"/\">Local Files</A> | \r\n");
		http_write_str(req, "<A HREF=\"/search/\">Refresh Response List</A> | \r\n");
		http_write_str(req, "<h1><a href=\"http://gnutella.wego.com/\">Gnutella</a> client web search</h1>");
		http_write_str(req, "<h2><a href=\"http://www.umr.edu/~jjp/\">Powered by Gnut v" VERSION "</a></h2>");
	}
}

void gnut_http_write_bottom(http_request_t *req)
{
	if (html_template_foot) {
		http_write_str(req,html_template_foot);
	} else {
		http_write_str(req, "</ol>");
		http_write_str(req, "</BODY>");
		http_write_str(req, "</HTML>\r\n");
	}
}

int gnut_http_write_form(http_request_t *req)
{
  g_debug(2,"entering\n");
  http_write_str(req, "<FORM ACTION=\"/search/\">");
  http_write_str(req, "Search Gnutella-net<p>");
  http_write_str(req, "<INPUT TYPE=\"text\" NAME=\"query\"><p>");

  http_write_str(req, "<INPUT TYPE=\"submit\"><p>");

#ifdef GNUT_HTTP_WAIT_FOR_RESULTS
  http_write_str(req, "Time to wait on req's");
  http_write_str(req, "<SELECT NAME=\"wait\">");
  http_write_str(req, "<OPTION VALUE=\"60\">60 sec</OPTION>");
  http_write_str(req, "<OPTION VALUE=\"30\">30 sec</OPTION>");
  http_write_str(req, "<OPTION VALUE=\"15\">15 sec</OPTION>");
  http_write_str(req, "</SELECT>");

  http_write_str(req, "Approx number of returns");
  http_write_str(req, "<SELECT NAME=\"returns\">");
  http_write_str(req, "<OPTION VALUE=\"20\">20</OPTION>");
  http_write_str(req, "<OPTION VALUE=\"40\">40</OPTION>");
  http_write_str(req, "<OPTION VALUE=\"60\">60</OPTION>");
  http_write_str(req, "<OPTION VALUE=\"80\">80</OPTION>");
  http_write_str(req, "<OPTION VALUE=\"100\">100</OPTION>");
  http_write_str(req, "</SELECT>");
#endif

  http_write_str(req, "</FORM>");

  return 0;
}

void gnut_http_template_scan() {
	FILE *fp;
	char *fname;
	char *new;
	long len;

	if (!conf_get_str("html_template")) return;
	if (strlen(conf_get_str("html_template"))<=1) return;
	fname=expand_path(conf_get_str("html_template"));
	g_debug(1, "path to html template: %s\n", fname);

    fp=fopen(fname,"r");
    if (fp==NULL) {
		g_debug(1, "Couldn't open html_template %s\n", fname);
		xfree(fname);
	} else {
		xfree(fname);

		// find the length
		fseek(fp,0,SEEK_END);
		len=ftell(fp);
		rewind(fp);
		g_debug(3, "html_template length: %i\n", len);

		// realloc space for the entire file
		new=realloc(html_template_head, len+1);
		if (!new) {
			g_debug(1, "Unable to realloc space for HTML template.\n");
			return;
		} 

		// slurp!
		fread(new, sizeof(char), len, fp);
		fclose(fp);
		html_template_head=new;

		// search for comment
		html_template_foot=strstr(new,"<!-- gnut -->");
		if (html_template_foot) {
			*html_template_foot = '\0';
			html_template_foot+=13;
		} else {
			g_debug(1, "Comment not found.\n");
			html_template_foot=&html_template_head[len];
		}
	}
}

#endif

static void gnut_guid_to_string(char *string, char *guid)
{
  int i;
  for (i = 0; i < 16; i++) {
   sprintf(&string[i*2], "%02x", (unsigned char)guid[i]);
  }
}

//static void gnut_string_to_guid(char *guid, char *string)
//{
//  int i;
//  for (i = 0; i < 16; i++) {
//    sscanf(&string[i*2], "%02x", (unsigned int) &guid[i]);
//  }
//}

