/* Jon Arney, (c) 2000 */

/* This file is distributed under the GPL, see file COPYING for details */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/socket.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#endif

#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <stdarg.h>

#include "gnut_lib.h"
#include "gnut_connection.h"
#include "gnut_net.h"
#include "host.h"
#include "gnut_transfer.h"
#include "query.h"
#include "conf.h"
#include "share.h"
#include "gnut_threads.h"
#include "gnut_http.h"
#include "Gnut_Hash.h"

http_request_t *http_request_new(int sock)
{
  http_request_t *req;

  g_debug(HTTP_REQUEST_DEBUG, "IN\n");
  req = malloc(sizeof(http_request_t));
  req->method = NULL;
  req->version = NULL;
  req->url_file = NULL;
  req->headers = NULL;
  req->response_headers = NULL;
  req->sock = sock;
  g_debug(HTTP_REQUEST_DEBUG, "OUT\n");
  return req;
}

void http_request_delete(http_request_t *req)
{
  g_debug(HTTP_REQUEST_DEBUG, "IN\n");
  if (req) {
    if (req->method) xfree(req->method);
    if (req->version) xfree(req->version);
    if (req->url_file) xfree(req->url_file);

    /*
     * HACK - slightly leaky - we should
     * walk the list and free the contents first.
     */
    if (req->headers) gnut_list_free(req->headers);
    if (req->response_headers) gnut_list_free(req->response_headers);
    xfree(req);
  }
  g_debug(HTTP_REQUEST_DEBUG, "OUT\n");
}

/*
 * Break the HTTP request into it's parts....
 */
int http_request_parse(http_request_t *req, char *request)
{
  char *ptr_to_http;
  char *ptr_to_file;
  int len;

  g_debug(HTTP_REQUEST_DEBUG, "IN\n");

  ptr_to_file = strstr(request, " ");
  if (!ptr_to_file) {
    g_debug(HTTP_REQUEST_DEBUG, "OUT - nothing to parse\n");
    return -1;
  }
  ptr_to_file++;

  ptr_to_http=strstr(request,"HTTP/");
  if (ptr_to_http==NULL) {
    g_debug(HTTP_REQUEST_DEBUG, "OUT - not an HTTP/1.0 or higher request\n");
    return -1;
  }

  /*
   * Save the "method" portion of the request.
   */
  len = (ptr_to_file - request);
  req->method = malloc(sizeof(char)*(len+1));
  strncpy(req->method, request, len);
  *(req->method+len) = '\0';

  /*
   * Save the "url" portion of the request.
   */
  len = (ptr_to_http-1-ptr_to_file);
  req->url_file = malloc(sizeof(char)*(len+1));
  strncpy(req->url_file, ptr_to_file, len);
  *(req->url_file+len) = '\0';

  /*
   * Save the HTTP version of the request - just in case we need it.
   */
  len = strlen(ptr_to_http);
  req->version = malloc(sizeof(char)*(len+1));
  strncpy(req->version, ptr_to_http, len);
  *(req->version + len) = '\0';

  g_debug(HTTP_REQUEST_DEBUG, "OUT\n");

  return 0;
}

/*
 * Examine the HTTP headers and find one of the values.
 */
char *http_request_find_header(http_request_t *req, char *name)
{
  Gnut_List *list;
  http_header_t *hdr;

  g_debug(HTTP_REQUEST_DEBUG, "IN - %s\n", name);

  list = req->headers;

  while (1) {
    if (!list) break;
    hdr = list->data;
    if (!hdr) break;

    if (!strcmp(hdr->name, name)) {
      g_debug(HTTP_REQUEST_DEBUG, "OUT - found header %s\n", hdr->value);
      return hdr->value;
    }

    list = gnut_list_next(list);
  }

  g_debug(HTTP_REQUEST_DEBUG, "OUT - no such header\n");

  return NULL;
}

/*
 * Read HTTP headers into the request object.
 */
int http_request_read_header(http_request_t *req)
{
  char buf[4096];
  int ret;
  http_header_t *hdr;

  g_debug(HTTP_REQUEST_DEBUG, "IN\n");

  /*
   * We must first read the remaining headers
   * and ignore them.
   */
  ret = 1;
  while (1) {
    errno = 0;
    ret = read_line(req->sock,buf,4096);
    if (errno == EINTR ) break;
    if (strlen(buf) < 3) break;

    hdr = http_header_new();
    if (!hdr) break;

    ret = http_header_parse(hdr, buf);
    if (ret) break;

    g_debug(1, "Got header %s:%s\n", hdr->name, hdr->value);
    g_debug(HTTP_REQUEST_DEBUG, "Got header \"%s: %s\"\n", hdr->name, hdr->value);

    req->headers = gnut_list_prepend(req->headers, hdr);
  }
  g_debug(HTTP_REQUEST_DEBUG, "OUT\n");
  return 0;
}

/*
 * Set the contents of an HTTP header.
 */
int http_response_header_add(http_request_t *req, char *name, char *value)
{
  http_header_t *hdr;

  g_debug(HTTP_HEADER_DEBUG, "IN - %s - %s\n", name, value);

  hdr = http_header_new();
  http_header_set(hdr, name, value);
  req->response_headers = gnut_list_prepend(req->response_headers, hdr);

  g_debug(HTTP_HEADER_DEBUG, "OUT\n");
  return 0;
}

int http_response_header_add_va(http_request_t *req, char *name, char *format, ...)
{
  va_list argp;
  char hdr_value[64];
  int ret;

  g_debug(HTTP_HEADER_DEBUG, "IN\n");

  va_start(argp,format);
  vsprintf(hdr_value,format,argp);
  va_end(argp);
  ret = http_response_header_add(req, name, hdr_value);

  g_debug(HTTP_HEADER_DEBUG, "OUT\n");
  return ret;
}

http_cookie_t *http_cookie_new(void)
{
  http_cookie_t *cookp;

  cookp = malloc(sizeof(http_cookie_t));
  cookp->value = NULL;
  cookp->path = NULL;
  cookp->domain = NULL;
  cookp->expiration = time(NULL);

  return cookp;
}

int http_header_get_cookie(http_request_t *req, http_cookie_t *cookp)
{
  char *end_of_value;
  int len_of_value;
  char *header_value;

  header_value = http_request_find_header(req, "Cookie");
  if (!header_value) return -1;

  end_of_value = strchr(header_value, ';');
  if (!end_of_value) {
    len_of_value = strlen(header_value)+1;
  }
  else {
    len_of_value = end_of_value - header_value;
  }

  cookp->value = malloc(sizeof(char)*len_of_value);
  if (!cookp->value) return -1;
  strncpy(cookp->value, header_value, len_of_value);
  cookp->value[len_of_value-1] = '\0';

  return 0;
}

void http_cookie_delete(http_cookie_t *cookie)
{
  if (cookie) {
    if (cookie->value) xfree(cookie->value);
    if (cookie->path) xfree(cookie->path);
    if (cookie->domain) xfree(cookie->domain);
    xfree(cookie);
  }
}

int http_response_set_cookie(http_request_t *req, http_cookie_t *cookie)
{
  return http_response_header_add_va(req, "Set-Cookie", "%s; path=%s", cookie->value, "/");
}

int http_response_header_write(http_request_t *req)
{
  http_header_t *hdr;
  Gnut_List *list;

  g_debug(HTTP_REQUEST_DEBUG, "IN\n");

  http_write_str(req, "HTTP/1.0 200 OK\r\n");

  list = req->response_headers;

  while (1) {
    if (!list) break;
    hdr = list->data;
    if (!hdr) break;
    http_write_va(req, HTTP_MAX_HEADER_SIZE, "%s: %s\r\n", hdr->name, hdr->value);
    list = gnut_list_next(list);
  }

  http_write_str(req, "\r\n");

  g_debug(HTTP_REQUEST_DEBUG, "OUT\n");

  return 0;
}

int http_write_str(http_request_t *req, char *a)
{
  int ret;

  g_debug(HTTP_IO_DEBUG, "IN %s\n", a);

  if (!req) return -1;
  if (req->sock < 0) return -1;
  ret = write(req->sock, a, strlen(a));

  g_debug(HTTP_IO_DEBUG, "OUT %i\n", ret);
  return ret;

}

int http_write_va(http_request_t *req, int maxlen, char *format, ...)
{
  va_list argp;
  int ret;
  char *value;

  g_debug(HTTP_IO_DEBUG, "IN\n");

  value = xmalloc(sizeof(char)*maxlen);

  va_start(argp,format);
  vsprintf(value,format,argp);
  va_end(argp);

  ret = http_write_str(req, value);

  xfree(value);

  g_debug(HTTP_IO_DEBUG, "OUT %i\n", ret);

  return ret;
}

